/* ----------------------------------------------------------------------
 * FILE: as504.y
 * PACKAGE: as504 - Siemens SAB-C504 Assembler.
 *
 * DESCRIPTION:
 *	This file contains the yacc parser for the assembler.
 *	Related to this are the following:
 *		error(), warning(), yyerror()
 *		genbyte(), genword(), genstr(), makeop()
 *
 * LICENCE:
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 * REVISION HISTORY:
 *	Jan. 19, 1990 - Created. (Ken Stauffer)
 *  Mar. 23, 2005 - Lots of additions, see http://www.vanwal.nl/as504/
 *                  (Freek van Walderveen)
 *
 * AUTHOR:
 *	The original author is Ken Stauffer (University of Calgary). Second
 *	author is Freek van Walderveen (email: freek -at- vanwal -dot- nl).
 *
 */

%{

#include <setjmp.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#define NOPE
#include "as504.h"
#undef NOPE

#define YYSTYPE union ystack
#define YYDEBUG 1	      // Generate debug code; needed for YYERROR_VERBOSE
#define YYERROR_VERBOSE // Give a more specific parse error message 

/* includes for D_EJECT */
#ifdef WIN32
#include <windows.h>
#include <winioctl.h>
#endif
#ifdef OS_linux
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/cdrom.h>
#endif

#define INCLUDE_STACK_SZ 	64	/* max num of nested INCLudes */
#define IF_STACK_SZ 	    64	/* max num of nested IFs */

void genbyte(int b);
void inclc(int i);
void dumplist(char *txt, int show);
void genstr(const char *s);
void genword(unsigned int w);
char *checkext(char *s, char *e);

void error(cs,a1,a2,a3,a4,a5,a6);
void warning(cs,a1,a2,a3,a4,a5,a6);
int makeop(struct opcode * op, struct mode *m, int add);
void yyerror (char *msg);
void emitbyte(int b);          // in emitter.c
void emitaddr(unsigned int a); // in emitter.c
int yylex(void);               // in lexer.c

extern int lineno;
extern int dashl;
extern char *asmfile;
extern jmp_buf main_env;
extern FILE *listing;
extern FILE *fin;
extern int debug;

int pass, fatal;
unsigned int lc;
int directive_was_org = 0, did_org = 0;

static unsigned char bytebuf[1024];		/* used by dumplist() */
static int bytecount;
int curregset = 0; /* current register set */
/* Include stack */
FILE *fp_stack[INCLUDE_STACK_SZ];
int linenumber_stack[INCLUDE_STACK_SZ];
char *filename_stack[INCLUDE_STACK_SZ];
int if_sp_stack[INCLUDE_STACK_SZ];
int include_sp = 0;
/* IF stack */
int if_expr_stack[IF_STACK_SZ];
int in_else_stack[IF_STACK_SZ];
int in_else = 0;
int if_expr = 1;
int if_sp = 0;
int trunc_len = 60;

/* ------------------------ G R A M M E R ----------------------------- */

%}

%token STRING
%token D_ORG
%token D_BYTE
/* %token D_DEFSTOR */
%token D_INCL
%token D_WORD
%token D_SKIP
%token D_USING
%token D_EQU
%token D_FLAG
%token D_IF
%token D_ELSE
%token D_ENDIF
%token D_WIDTH
%token D_NOP
%token D_EJECT
%token D_TOLST
%token D_TOCON
%token ACALL
%token ADD
%token ADDC
%token AJMP
%token ANL
%token AR0
%token AR1
%token AR2
%token AR3
%token AR4
%token AR5
%token AR6
%token AR7
%token CJNE
%token CLR
%token CPL
%token DA
%token DEC
%token DIV
%token DJNZ
%token INC
%token JB
%token JBC
%token JC
%token JMP
%token JNB
%token JNC
%token JNZ
%token JZ
%token LCALL
%token LJMP
%token MOV
%token MOVC
%token MOVX
%token NOP
%token MUL
%token ORL
%token POP
%token PUSH
%token RET
%token RETI
%token RL
%token RLC
%token RR
%token RRC
%token SETB
%token SJMP
%token SUBB
%token SWAP
%token XCH
%token XCHD
%token XRL
%token AB
%token A
%token C
%token PC
%token DPTR
%token BITPOS
%token R0
%token R1
%token R2
%token R3
%token R4
%token R5
%token R6
%token R7
%token VALUE
%token SYMBOL
%token HIGH
%token LOW
%token NOT
%token BITNOT
%token AND
%token OR
%token XOR
%token SHL
%token SHR
%token MOD

%left OR XOR
%left AND
%left '+' '-'
%left '*' '/' MOD SHL SHR
%left NOT BITNOT

%start program

%%
program		:	linelist
{
}
		;

linelist	: linelist line { directive_was_org = 0; }
		| line { directive_was_org = 0; }
		;

line		: undefsym ':' linerest
{
  if (if_expr ^ in_else) {
	if( pass1 ) {
		$1.sym->type = LABEL;
		$1.sym->value = lc;
      if (debug) {
        printf("Pass 1, value=%d\n", lc);
      }
      if (directive_was_org) {
        fprintf(stderr,"File: %s, line: %d, warning: label in front of ORG directive, your code will probably work with this compiler, but ASM504 will first define the label \n",asmfile,lineno);
        directive_was_org = 0;
      }
	}
	inclc($3.value);
  }
	bytecount = 0;
}
		| linerest		{ if (if_expr ^ in_else) { inclc($1.value); } bytecount = 0; }
    | specdirective { if (if_expr ^ in_else) { inclc($1.value); } bytecount = 0; }
		;

linerest	: directive '\n'	{
						$$.value = $1.value;
						if( dashl && pass2 )
							dumplist($2.str,1);
					}
		| instr '\n'		{
						$$.value = $1.value;
						if( dashl && pass2 )
							dumplist($2.str,1);

					}
		| '\n'			{
						$$.value = 0;
						if( dashl && pass2 )
							dumplist($1.str,0);
					}
		;

/* --------------------
 * DIRECTIVES:
 *
 */

dotornot : '.' | ;

directive	: dotornot D_ORG defexpr
{
	lc = $3.val.v;
	if( pass2 ) emitaddr(lc);
	bytecount = 0;
	$$.value = 0;
  directive_was_org = 1;
  did_org = 1;
}
		| dotornot D_BYTE blist	{ $$.value = $3.value; }
    /*
		| D_DEFSTOR data8	{
                        $$.value = $2.value;
                        if (pass2) {
                          int i;
                          for (i = 0; i < $2.value; i++) {
                            genbyte(0);
                          }
                        }
                      }
    */
		| dotornot D_WORD wlist	{ $$.value = $3.value; }
		| dotornot D_SKIP defexpr	{ $$.value = $3.val.v;
					  if( pass2 )
						emitaddr(lc+$$.value); }
		| dotornot D_EQU undefsym ',' expr
{
	if( $5.val.d == 0 )
		error("Expression is undefined in pass 1");
	$3.sym->type = LABEL;
	$3.sym->value = $5.val.v;
	$$.value = 0;
}
		| undefsym dotornot D_EQU expr
{
	if( $4.val.d == 0 )
		error("Expression is undefined in pass 1");
	$1.sym->type = LABEL;
	$1.sym->value = $4.val.v;
	$$.value = 0;
}
		| dotornot D_FLAG SYMBOL ',' flag
{
	$3.sym->type = LABEL;
	$3.sym->value = $5.value;
	$$.value = 0;
}
		| dotornot D_NOP			      { $$.value = 0; }
    | dotornot D_NOP noplist   { $$.value = 0; }
		| dotornot D_USING defexpr
{
  if ($3.value < 0 || $3.value > 3) {
    error("register set can only be 0, 1, 2 or 3");
  }
  curregset = $3.value;
  $$.value = 0;
}
		;

noplist : defexpr   { $$.value = 0; }
        | defexpr ',' noplist   { $$.value = 0; }
        ;

specdirective : dotornot D_INCL STRING '\n'
{
  char *pc;
  char *filename;
  int i, j, fnlen;
  
  if (if_expr ^ in_else) {
    /* check for stack overflow */
    if (include_sp == INCLUDE_STACK_SZ) {
      error("too many nested includes");
    }

    if (debug) {
      printf("Pushing includefile ``%s''\n", $3.str);
    }

    /* save the old file pointer, filename, if_sp and linenumber on the stack */
    fp_stack[include_sp] = fin;
    (void)strcpy((pc = (char *)malloc((unsigned)strlen(asmfile) + 1)), asmfile);
    filename_stack[include_sp] = pc;
    if_sp_stack[include_sp] = if_sp;
    linenumber_stack[include_sp++] = lineno;

    /* try opening the file */
    filename = $3.str;
    if (checkext(filename, ".asm") == NULL) {
      if (debug) {
        printf("Adding .asm extension to included filename.\n");
      }
		  /* add .asm extension */
      fnlen = strlen(filename);
      filename = (char *)realloc(filename, fnlen + 5);
      filename[fnlen++] = '.';
      filename[fnlen++] = 'a';
      filename[fnlen++] = 's';
      filename[fnlen++] = 'm';
      filename[fnlen] = '\0';
    }
    if ((fin = fopen(filename,"r")) == NULL) {
      /* try adding current working dir. */
      for (i = strlen(asmfile); i >= 0; i--) {
#ifdef WIN32
        if (asmfile[i] == '\\') {
          break;
        }
#else
        if (asmfile[i] == '/') {
          break;
        }
#endif
      }
      fnlen = strlen(filename);
      filename = (char *)realloc(filename, i + fnlen + 3);
      for (j = i + fnlen + 2; j > i; j--) {
        filename[j] = filename[j - i - 1];
      }
      for (j = 0; j <= i; j++) {
        filename[j] = asmfile[j];
      }
      if ((fin = fopen(filename, "r")) == NULL) {
        error("cannot open included file: ``%s''", filename);
      }
	  }

    /* reset line number and filename */
	  lineno = 0;
    free(asmfile);
    asmfile = filename;
  }
  
	$$.value = 0;
}
              | dotornot D_IF defexpr '\n'
{
  /* check for stack overflow */
  if (if_sp == IF_STACK_SZ) {
    error("too many nested ifs");
  }
  
  /* push current if_expr and in_else on the stack */
  if_expr_stack[if_sp] = if_expr;
  in_else_stack[if_sp] = in_else;
  if_sp++;
  
  /* update if_expr and in_else */
  if_expr = ($3.value != 0);
  if (debug) {
    printf("IF expression evaluates to: %s\n", (if_expr?"TRUE":"FALSE"));
  }
  in_else = 0;

  $$.value = 0;
}
              | dotornot D_ELSE '\n'
{
  /* check if there was an starting-if statement in this file */
  if (include_sp == 0) {
    /* in main file */
    if (if_sp == 0) {
      error("ELSE without IF");
    }
  } else {
    /* in some included file */
    if (if_sp == if_sp_stack[include_sp-1]) {
      error("ELSE without IF");
    }
  }
  
  /* check if there wasn't already an ELSE in this IF */
  if (in_else) {
    error("second ELSE in this IF block");
  }
  
  /* update in_else */
  in_else++;

  $$.value = 0;
}
              | dotornot D_ENDIF '\n'
{
  /* check if there was an starting-if statement in this file */
  if (include_sp == 0) {
    /* in main file */
    if (if_sp == 0) {
      error("ENDIF without IF");
    }
  } else {
    /* in some included file */
    if (if_sp == if_sp_stack[include_sp-1]) {
      error("ENDIF without IF");
    }
  }
  
  if_sp--;
  
  /* pop if_expr and in_else */
  if_expr = if_expr_stack[if_sp];
  in_else = in_else_stack[if_sp];

  $$.value = 0;
}
              | dotornot D_WIDTH defexpr '\n'
{
  if (if_expr ^ in_else) {
    if ($3.value > 255 || $3.value < 19) {
      error("incorrect WIDTH");
    }
    trunc_len = $3.value - 19;
  }
  $$.value = 0;
}
              | dotornot D_EJECT '\n'
{
/* I'm getting tired of trying to be compatible with all those useless
   ASM504 compiler directives. I'll nevertheless accept an EJECT statement,
   my own way that is... */

 if (if_expr ^ in_else) { 

#if (defined(OS_openbsd) || defined(OS_freebsd) || defined(OS_netbsd))
    int fd;
    if ((fd = open("/dev/cdrom", O_RDONLY | O_NONBLOCK)) >= 0) {
      /* printf("Trying to eject /dev/cdrom...\n"); */
      (void) ioctl (fd, CDIOCALLOW);
      ioctl (fd, CDIOCEJECT);
    }
#endif

#ifdef WIN32
    CHAR lpszDrive[4];
    char cDrive;
    for (cDrive='A'; cDrive < 'Z'; cDrive++) {
      sprintf(lpszDrive,"%c:\\",cDrive);

      //gets the drive type, if its a CD-ROM, try to eject it
      if (GetDriveType(lpszDrive) == DRIVE_CDROM) {
        /* printf("%s is the first cdrom drive, trying to eject it...\n", lpszDrive); */
        HANDLE hFile;
        CHAR lpszDeviceName[7];
        /* take the drive letter and put it in the format used in CreateFile */
        sprintf(lpszDeviceName, "\\\\.\\%c:", *lpszDrive);
        hFile = CreateFile(lpszDeviceName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, NULL); 
        DWORD lpBytesReturned;
        DeviceIoControl(hFile, IOCTL_STORAGE_EJECT_MEDIA, NULL, 0, NULL, 0, &lpBytesReturned, NULL);
        break;
      }
    } 
#endif // WIN32

#ifdef OS_linux
    int fd;
    if ((fd = open("/dev/cdrom", O_RDONLY | O_NONBLOCK)) >= 0) {
      /* printf("Trying to eject /dev/cdrom...\n"); */
      ioctl(fd, CDROMEJECT, 0);
    }
#endif
  }
 
  $$.value = 0;
}
              | dotornot D_TOLST tolstlist '\n' { $$.value = 0; }
              | dotornot D_TOCON toconlist '\n' { $$.value = 0; }
              ;

tolstlist : STRING  { if (dashl && pass2 && (if_expr ^ in_else)) { fprintf(listing, "%s\n", $1.str); } $$.value = 0; free($1.str); }
          | defexpr { if (dashl && pass2 && (if_expr ^ in_else)) { fprintf(listing, "%d\n", $1.value); } $$.value = 0; }
          | tolstlist ',' STRING  { if (dashl && pass2 && (if_expr ^ in_else)) { fprintf(listing, "%s\n", $3.str); } $$.value = 0; free($3.str); }
          | tolstlist ',' defexpr { if (dashl && pass2 && (if_expr ^ in_else)) { fprintf(listing, "%d\n", $3.value); } $$.value = 0; }
          ;
          
toconlist : STRING  { if (pass2 && (if_expr ^ in_else)) { printf("%s\n", $1.str); } $$.value = 0; free($1.str); }
          | defexpr { if (pass2 && (if_expr ^ in_else)) { printf("%d\n", $1.value); } $$.value = 0; }
          | toconlist ',' STRING  { if (pass2 && (if_expr ^ in_else)) { printf("%s\n", $3.str); } $$.value = 0; free($3.str); }
          | toconlist ',' defexpr { if (pass2 && (if_expr ^ in_else)) { printf("%d\n", $3.value); } $$.value = 0; }
		;

defexpr		: expr
{
		if( $1.val.d == 0 )
			error("expression is undefined in pass 1");
		if( !(isbit16($1.val.v)) )
			error("value greater than 16-bits");
		$$.value = $1.val.v;
}
		;

flag		: flagv BITPOS
{
	if( !isbit8($1.value) )
		warning("bit address exceeds 8-bits");
	if( isbmram($1.value) )
		$$.value = ($1.value-0x20)*8+ $2.value;
	else if( isbmsfr($1.value) )
		$$.value = $1.value + $2.value;
	else
		warning("invalid bit addressable RAM location");
}
		;

flagv		: SYMBOL
{
	if( $1.sym->type == UNDEF )
		error("Symbol %s must be defined in pass 1",$1.sym->name);
	$$.value = $1.sym->value;
}
		| VALUE			{ $$.value = $1.value; }
		;


undefsym	: SYMBOL
{
  if (if_expr ^ in_else) {
	if( $1.sym->type != UNDEF && pass1)
		  error("Attempt to redefine symbol: %s",$1.sym->name);
    if (debug) {
      printf("Symbol found: %s\n", $1.sym->name);
    }
	$$.sym = $1.sym;
  }
}
		;

blist		: blist ',' data8
{
	if( pass2 ) genbyte($3.value);
	$$.value = $1.value + 1;
}
		| blist ',' STRING
{
		$$.value = $1.value + strlen($3.str);
	if (!pass1) {
		genstr($3.str);
	}
	free($3.str);
}
		| data8
{
	if( pass2 ) genbyte($1.value);
	$$.value = 1;
}
		| STRING
{
		$$.value = 0; /* = strlen($1.str); */
    $$.str = $1.str;
    /*
	if (!pass1) {
		genstr($1.str);
	}
	free($1.str);
  */
}
		;

wlist		: wlist ',' data16
{
	if( pass2 ) genword($3.value);
	$$.value = $1.value + 2;
}
		| data16
{
	if( pass2 ) genword($1.value);
	$$.value = 2;
}
		;



/* --------------------
 * EXPRESSIONS:
 *
 */

expr		: '*'     { $$.val.v = lc; $$.val.d = 1; }

        | '$'     { $$.val.v = lc; $$.val.d = 1; }
        
        | '(' expr ')'  { $$.val.v = $2.val.v;
                          $$.val.d = $2.val.d; }

        | '-' expr %prec '*'	{ $$.val.v = -$2.val.v;
  			        $$.val.d = $2.val.d;  }

        | NOT expr		    { $$.val.v = !($2.val.v);
  			        $$.val.d = $2.val.d; }

        | BITNOT expr		    { $$.val.v = ~($2.val.v);
  			        $$.val.d = $2.val.d; }

        | expr OR expr		{ $$.val.v = $1.val.v | $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr XOR expr		{ $$.val.v = $1.val.v ^ $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr AND expr		{ $$.val.v = $1.val.v & $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr '*' expr		{ $$.val.v = $1.val.v * $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr '/' expr		{ $$.val.v = $1.val.v / $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr MOD expr		{ $$.val.v = $1.val.v % $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr SHL expr		{ $$.val.v = $1.val.v << $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr SHR expr		{ $$.val.v = $1.val.v >> $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr '-' expr		{ $$.val.v = $1.val.v - $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }

        | expr '+' expr		{ $$.val.v = $1.val.v + $3.val.v;
  			        $$.val.d = $1.val.d && $3.val.d; }
        | SYMBOL
              {
	              if( pass1 ) {
		              $$.val.v = $1.sym->value;
		              $$.val.d = ($1.sym->type != UNDEF);
	              }
	              else {
		              if( $1.sym->type == UNDEF )
			              error("Undefined symbol %s",$1.sym->name);
		              $$.val.v = $1.sym->value;
		              $$.val.d = 1;
	              }
              }
        | VALUE { $$.val.v = $1.val.v; $$.val.d=1; }
		;





/* --------------------
 * INSTRUCTIONS:
 *
 */

instr		: NOP
				{ $$.value = makeop($1.op,NULL,0); }
		| ACALL addr11
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| AJMP addr11
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| ADD two_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| ADDC two_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| SUBB two_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| XRL two_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| XRL two_op2
				{ $$.value = makeop($1.op,&$2.mode,4); }
		| ANL two_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| ANL two_op2
				{ $$.value = makeop($1.op,&$2.mode,4); }
		| ANL two_op3
				{ $$.value = makeop($1.op,&$2.mode,6); }
		| ORL two_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| ORL two_op2
				{ $$.value = makeop($1.op,&$2.mode,4); }
		| ORL two_op3
				{ $$.value = makeop($1.op,&$2.mode,6); }
		| XCH two_op1
				{ if( get_md($2.mode) == 3 )
					error("Immediate mode is illegal");
				  $$.value = makeop($1.op,&$2.mode,0);
				}
		| INC single_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| INC DPTR
				{ $$.value = makeop($1.op,NULL,4); }
		| DEC single_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| DA A
				{ $$.value = makeop($1.op,NULL,0); }
		| DIV AB
				{ $$.value = makeop($1.op,NULL,0); }
		| JMP '@' A '+' DPTR
				{ $$.value = makeop($1.op,NULL,0); }
		| JMP '@' DPTR '+' A
				{ $$.value = makeop($1.op,NULL,0); }
		| MUL AB
				{ $$.value = makeop($1.op,NULL,0); }
		| RET
				{ $$.value = makeop($1.op,NULL,0); }
		| RETI
				{ $$.value = makeop($1.op,NULL,0); }
		| RL A
				{ $$.value = makeop($1.op,NULL,0); }
		| RLC A
				{ $$.value = makeop($1.op,NULL,0); }
		| RR A
				{ $$.value = makeop($1.op,NULL,0); }
		| RRC A
				{ $$.value = makeop($1.op,NULL,0); }
		| SWAP A
				{ $$.value = makeop($1.op,NULL,0); }
		| XCHD two_op1
				{ if( get_md($2.mode) != 2 )
					error("Invalid addressing mode");
				  $$.value = makeop($1.op,&$2.mode,-2); }
		| CLR single_op2
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| CPL single_op2
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| SETB single_op2
				{ if( get_md($2.mode) == 0 )
					error("Invalid addressing mode");
				  $$.value = makeop($1.op,&$2.mode,-1); }
		| PUSH data8
				{
				   struct mode tmp;
					set_md(tmp,0);
					set_ov(tmp,0);
					set_sz(tmp,1);
					set_b1(tmp,$2.value);
					$$.value = makeop($1.op,&tmp,0);
				}
		| POP data8
				{
				   struct mode tmp;
					set_md(tmp,0);
					set_ov(tmp,0);
					set_sz(tmp,1);
					set_b1(tmp,$2.value);
					$$.value = makeop($1.op,&tmp,0);
				}
		| LJMP addr16
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| LCALL addr16
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| JC relative
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| JNC relative
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| JNZ relative
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| JZ relative
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| SJMP relative
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| CJNE three_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| JB two_op4
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| JNB two_op4
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| JBC two_op4
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| DJNZ two_op5
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| MOV two_op1
				{ $$.value = makeop($1.op,&$2.mode,0); }
		| MOV two_op2
				{ $$.value = makeop($1.op,&$2.mode,4); }
		| MOV two_op6
				{ $$.value = makeop($1.op,&$2.mode,6); }


		| MOVC A ',' '@' A '+' DPTR
				{ $$.value = makeop($1.op,NULL,0); }
		| MOVC A ',' '@' DPTR '+' A
				{ $$.value = makeop($1.op,NULL,0); }
		| MOVC A ',' '@' A '+' PC
				{ $$.value = makeop($1.op,NULL,1); }
		| MOVC A ',' '@' PC '+' A
				{ $$.value = makeop($1.op,NULL,1); }

		| MOVX A ',' '@' regi
				{ $$.value = makeop($1.op,NULL,$5.value); if(pass1) fprintf(stderr,"File: %s, line: %d, warning: MOVX A, @Ri should only be used for systems with a maximum of 256 bytes of external data memory!\n",asmfile,lineno); }
		| MOVX A ',' '@' DPTR
				{ $$.value = makeop($1.op,NULL,2); }
		| MOVX '@' regi ',' A
				{ $$.value = makeop($1.op,NULL,$3.value+3);  if(pass1) fprintf(stderr,"File: %s, line: %d, warning: MOVX @Ri, A should only be used for systems with a maximum of 256 bytes of external data memory!\n",asmfile,lineno); }
		| MOVX '@' DPTR ',' A
				{ $$.value = makeop($1.op,NULL,5); }
		;




/* --------------------
 * ADDRESSING MODES:
 *
 */

two_op1		: A ',' reg
				{
					set_md($$.mode,0);
					set_ov($$.mode, $3.value);
					set_sz($$.mode, 0);
				}
		| A ',' data8
				{
					set_md($$.mode,1);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$3.value);
				}
		| A ',' '@' regi
				{
					set_md($$.mode,2);
					set_ov($$.mode,$4.value);
					set_sz($$.mode,0);
				}
		| A ',' '#' data8
				{
					set_md($$.mode,3);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$4.value);
				}
		;

two_op2		: data8 ',' A
				{
					set_md($$.mode,0);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$1.value);
				}
		| data8 ',' '#' data8
				{
					set_md($$.mode,1);
					set_ov($$.mode,0);
					set_sz($$.mode,2);
					set_b1($$.mode,$1.value);
					set_b2($$.mode,$4.value);
				}
		;

two_op3		: C ',' bit
				{
					set_md($$.mode,0);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$3.value);
				}
		| C ',' '/' bit
				{
					set_md($$.mode,1);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$4.value);
				}
		| C ',' '!' bit
				{
					set_md($$.mode,1);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$4.value);
				}
		;

two_op4		: bit ',' rel
				{
					set_md($$.mode,0);
					set_ov($$.mode,0);
					set_sz($$.mode,2);
					set_b1($$.mode,$1.value);
					set_b2($$.mode,$3.value);
				}
		;

two_op5		: reg ',' rel2
				{
					set_md($$.mode,0);
					set_ov($$.mode,$1.value);
					set_sz($$.mode,1);
					set_b1($$.mode,$3.value);
				}
		| data8 ',' rel
				{
					set_md($$.mode,1);
					set_ov($$.mode,0);
					set_sz($$.mode,2);
					set_b1($$.mode,$1.value);
					set_b2($$.mode,$3.value);
				}
		;

two_op6		: reg ',' A
				{
					set_md($$.mode,0);
					set_ov($$.mode,$1.value);
					set_sz($$.mode,0);
				}
		| reg ',' data8
				{
					set_md($$.mode,1);
					set_ov($$.mode,$1.value);
					set_sz($$.mode,1);
					set_b1($$.mode,$3.value);
				}
		| reg ',' '#' data8
				{
					set_md($$.mode,2);
					set_ov($$.mode,$1.value);
					set_sz($$.mode,1);
					set_b1($$.mode,$4.value);
				}
		| data8 ',' reg
				{
					set_md($$.mode,3);
					set_ov($$.mode,$3.value);
					set_sz($$.mode,1);
					set_b1($$.mode,$1.value);
				}
		| data8 ',' data8
				{
					set_md($$.mode,4);
					set_ov($$.mode,0);
					set_sz($$.mode,2);
					set_b1($$.mode,$3.value);
					set_b2($$.mode,$1.value);
				}
		| data8 ',' '@' regi
				{
					set_md($$.mode,5);
					set_ov($$.mode,$4.value);
					set_sz($$.mode,1);
					set_b1($$.mode,$1.value);
				}
		| '@' regi ',' A
				{
					set_md($$.mode,6);
					set_ov($$.mode,$2.value);
					set_sz($$.mode,0);
				}
		| '@' regi ',' data8
				{
					set_md($$.mode,7);
					set_ov($$.mode,$2.value);
					set_sz($$.mode,1);
					set_b1($$.mode,$4.value);
				}
		| '@' regi ',' '#' data8
				{
					set_md($$.mode,8);
					set_ov($$.mode,$2.value);
					set_sz($$.mode,1);
					set_b1($$.mode,$5.value);
				}
		| DPTR ',' '#' data16
			{
				set_md($$.mode,9);
				set_ov($$.mode,0);
				set_sz($$.mode,2);
				set_b1($$.mode, ($4.value & 0xff00) >> 8 );
				set_b2($$.mode, ($4.value & 0x00ff) );
			}
		| C ',' bit
				{
					set_md($$.mode,10);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$3.value);
				}
	/*
	 * Following two productions cannot be represented by:
	 *
	 *	bit ',' C
	 *
	 * Because yacc gives tons of reduce/reduce errors if
 	 * that is attempted.
	 *
	 */
		| data8 ',' C
				{
					set_md($$.mode,11);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$1.value);
				}
		| data8 BITPOS ',' C
{
	if( pass2 ) {
		if( !isbit8($1.value) )
			warning("Bit address exceeds 8-bits");
		if( isbmram($1.value) )
			set_b1($$.mode, ($1.value-0x20)*8+ $2.value );
		else if( isbmsfr($1.value) )
			set_b1($$.mode, $1.value + $2.value );
		else
			warning("Invalid bit addressable RAM location");
	}
	set_md($$.mode,11);
	set_ov($$.mode,0);
	set_sz($$.mode,1);
}
		;


single_op1	: A
				{
					set_md($$.mode,0);
					set_ov($$.mode,0);
					set_sz($$.mode,0);
				}

		| reg
				{
					set_md($$.mode,1);
					set_ov($$.mode,$1.value);
					set_sz($$.mode,0);
				}
		| data8
				{
					set_md($$.mode,2);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$1.value);
				}
		| '@' regi
				{
					set_md($$.mode,3);
					set_ov($$.mode,$2.value);
					set_sz($$.mode,0);
				}
		;

single_op2	: A
				{
					set_md($$.mode,0);
					set_ov($$.mode,0);
					set_sz($$.mode,0);
				}
		| C
				{
					set_md($$.mode,1);
					set_ov($$.mode,0);
					set_sz($$.mode,0);
				}
		| bit
				{
					set_md($$.mode,2);
					set_ov($$.mode,0);
					set_sz($$.mode,1);
					set_b1($$.mode,$1.value);
				}
		;

three_op1	: A ',' data8 ',' rel
				{
					set_md($$.mode,0);
					set_ov($$.mode,0);
					set_sz($$.mode,2);
					set_b1($$.mode,$3.value);
					set_b2($$.mode,$5.value);
				}
		| A ',' '#' data8 ',' rel
				{
					set_md($$.mode,1);
					set_ov($$.mode,0);
					set_sz($$.mode,2);
					set_b1($$.mode,$4.value);
					set_b2($$.mode,$6.value);
				}
		| reg ',' '#' data8 ',' rel
				{
					set_md($$.mode,2);
					set_ov($$.mode,$1.value);
					set_sz($$.mode,2);
					set_b1($$.mode,$4.value);
					set_b2($$.mode,$6.value);
				}
		| '@' regi ',' '#' data8 ',' rel
				{
					set_md($$.mode,3);
					set_ov($$.mode,$2.value);
					set_sz($$.mode,2);
					set_b1($$.mode,$5.value);
					set_b2($$.mode,$7.value);
				}
		;

rel		: expr
{
		int offset;
		if( pass2 ) {
			offset = $1.val.v - (lc+3);
			if( offset > 127 || offset < -128 )
			   warning("Relative offset exceeds -128 / +127");
			$$.value = offset;
		}
}
		;

/*
 * This production differs from the above, by 1 number!
 *
 */

rel2		: expr
{
		int offset;
		if( pass2 ) {
			offset = $1.val.v - (lc+2); /* different! */
			if( offset > 127 || offset < -128 )
			   warning("Relative offset exceeds -128 / +127");
			$$.value = offset;
		}
}
		;


bit		: bitv BITPOS
{
	if( pass2 ) {
		if( !isbit8($1.value) )
			warning("Bit address exceeds 8-bits");
		if( isbmram($1.value) )
			$$.value = ($1.value-0x20)*8+$2.value;
		else if( isbmsfr($1.value) )
			$$.value = $1.value + $2.value;
		else
			warning("Invalid bit addressable RAM location");
	}
}
		| bitv
{
	if( pass2 ) {
		if( !isbit8($1.value) )
			warning("Bit address exceeds 8-bits");
		$$.value = $1.value;
	}
}
		;

bitv		: SYMBOL
{
	if( $1.sym->type == UNDEF && pass2 )
		error("Symbol %s undefined",$1.sym->name);
	$$.value = $1.sym->value;
}
		| VALUE		{ $$.value = $1.value; }
		;

reg		: R0		{ $$.value = 0; }
		| R1		{ $$.value = 1; }
		| R2		{ $$.value = 2; }
		| R3		{ $$.value = 3; }
		| R4		{ $$.value = 4; }
		| R5		{ $$.value = 5; }
		| R6		{ $$.value = 6; }
		| R7		{ $$.value = 7; }
		;

regi		: R0		{ $$.value = 0; }
		| R1		{ $$.value = 1; }
		| R2
				{ $$.value = 0;
				  warning("Illegal indirect register: @r2"); }
		| R3
				{ $$.value = 0;
				  warning("Illegal indirect register: @r3"); }
		| R4
				{ $$.value = 0;
				  warning("Illegal indirect register: @r4"); }
		| R5
				{ $$.value = 0;
				  warning("Illegal indirect register: @r5"); }
		| R6
				{ $$.value = 0;
				  warning("Illegal indirect register: @r6"); }
		| R7
				{ $$.value = 0;
				  warning("Illegal indirect register: @r7"); }
		;

areg  : AR0		{ $$.value = 0 + curregset * 8; }
		  | AR1		{ $$.value = 1 + curregset * 8; }
		  | AR2		{ $$.value = 2 + curregset * 8; }
  		| AR3		{ $$.value = 3 + curregset * 8; }
  		| AR4		{ $$.value = 4 + curregset * 8; }
  		| AR5		{ $$.value = 5 + curregset * 8; }
  		| AR6		{ $$.value = 6 + curregset * 8; }
  		| AR7		{ $$.value = 7 + curregset * 8; }
  		;

data8   : HIGH  '(' data16 ')'  { $$.value = ($3.value & 0xff00) >> 8; }
        | LOW   '(' data16 ')'  { $$.value = $3.value & 0xff; }
        | realdata8             { $$.value = $1.value; }
        | areg                  { $$.value = $1.value; }
        ;

realdata8		: expr
{
	if( pass2 ) {
		if( !isbit8($1.val.v) )
			warning("Expression greater than 8-bits");
	}
	$$.value = $1.val.v;
}
		;

data16		: expr
{
	if( pass2 ) {
		if( !isbit16($1.val.v) )
			warning("Expression greater than 16-bits");
	}
	$$.value = $1.val.v;
}
		;

addr11		: expr
{
		if( pass2 ) {
			if( !isbit16($1.val.v)  )
				warning("Address greater than 16-bits");
			if( ($1.val.v & size11) != ((lc+2) & size11) )
				warning("Address outside current 2K page");
		}
		set_md($$.mode,0);
		set_ov($$.mode, ($1.val.v&0x0700)>>3 );
		set_sz($$.mode,1);
		set_b1($$.mode,$1.val.v&0x00ff);
}
		;

addr16		: expr
{
		if( pass2 ) {
			if( !isbit16($1.val.v)  )
				warning("Address greater than 16-bits");
		}
		set_md($$.mode,0);
		set_ov($$.mode, 0 );
		set_sz($$.mode,2);
		set_b1($$.mode, ($1.val.v & 0xff00 ) >> 8 );
		set_b2($$.mode, ($1.val.v & 0x00ff ) );
}
		;

relative  : expr
{
		int offset=0;
		if( pass2 ) {
			offset = $1.val.v - (lc+2);
			if( offset>127 || offset<-128 )
			   warning("Relative offset exceeds -128 / +127");
		}
		set_md($$.mode,0);
		set_ov($$.mode,0);
		set_sz($$.mode,1);
		set_b1($$.mode,offset);
}         
          ;
/*
          | '$'
{
		int offset=0;
		if (pass2) {
			offset = -2;
		}
		set_md($$.mode,0);
		set_ov($$.mode,0);
		set_sz($$.mode,1);
		set_b1($$.mode,offset);
}
          ;
*/
%%

/* ---------------------------------------------------------------------- */

void yyerror(char *s)
{
	error(s);
}


/* ----------------------------------------------------------------------
 * error:
 *	Uses semi-variable arguments. This causes immediate assembler
 *	termination.
 */

void error(cs,a1,a2,a3,a4,a5,a6)
char *cs,*a1,*a2,*a3,*a4,*a5,*a6;
{
  int i;
  
	fprintf(stderr,"File: %s, line: %d, ",asmfile,lineno);
	fprintf(stderr,cs,a1,a2,a3,a4,a5,a6);
	fprintf(stderr,".\n");
  for (i = 0; i < include_sp; i++) {
    fprintf(stderr,"Included from %s, line: %d\n", filename_stack[i], linenumber_stack[i]);
  }
	longjmp(main_env,1);
}

/* ----------------------------------------------------------------------
 * warning:
 *	Produce error message. This will abort assembly at
 *	the end of the current pass.
 *
 */

void warning(cs,a1,a2,a3,a4,a5,a6)
char *cs,*a1,*a2,*a3,*a4,*a5,*a6;
{
	fatal++;
	fprintf(stderr,"File: %s, line: %d, ",asmfile,lineno);
	fprintf(stderr,cs,a1,a2,a3,a4,a5,a6);
	fprintf(stderr,".\n");
}


/* ----------------------------------------------------------------------
 * makeop:
 *	This function makes an opcode based on the instruction symbol table
 *	entry, and an addressing mode structure.
 *	This function is called from both passes, but
 *	only generates code in pass 2.
 *
 *	Resultant opcode bytes are passed to genbyte().
 *
 *	Returns the number of bytes that the instruction
 *	occupies.
 *
 */

int makeop(struct opcode * op, struct mode *m, int add)
{
	register unsigned int newop;

  if (if_expr ^ in_else) {
	  if( m == NULL ) {
		  if(pass2) genbyte(op->bytes[0+add]);
		  return(1);
	  }

	  if( pass2 ) {
		  newop = op->bytes[ get_md(*m)+add ] | get_ov(*m);
		  genbyte(newop);
		  if( get_sz(*m) > 0 ) genbyte( get_b1(*m) );
		  if( get_sz(*m) > 1 ) genbyte( get_b2(*m) );
	  }
	  return( get_sz(*m)+1 );
  } else {
    return 0;
  }
}


/* ----------------------------------------------------------------------
 * inclc:
 *	Increments the Location Counter by 'i' amount.
 *	Check to see if 'i' overflows 64K.
 *	Checks to see if assembler is overlapping previous sections
 *	of code. (using a large bit field).
 *
 */

#define indx(a) ( (a)/(sizeof(int)*8) )
#define bit(a)	( 1 << ((a)%(sizeof(int)*8)) )

#define getloc(a) (regions[indx(a)] & bit(a))
#define setloc(a) (regions[indx(a)] |= bit(a))

static unsigned int regions[ 0x10000/(sizeof(int)*8) ];
void inclc(int i)
{

  if (if_expr ^ in_else) {
	  while(i-- > 0) {
		if( pass2 && getloc(lc) )
			  error("Location counter overlaps (you're trying to overwrite a byte that was already filled, perhaps you did ORG twice to the same location)");
		if( pass2 ) setloc(lc);
		lc += 1;
	}
		
	if( lc > 0xffff )
		error("Location counter has exceeded 16-bits");
  }
}

/* ----------------------------------------------------------------------
 * padline:
 *	This routine returns a new string, which is equivilant to
 *	'line' except that all tabs have been expanded to spaces, and
 *	the total length has been truncated to trunc_len chars.
 */

char *padline(char *line)
{
	char *newline = (char *)malloc(trunc_len + 1);
	char *p1;
	int pos=0,nxtpos;

	for(p1=line; pos < trunc_len && *p1; p1++ ) {
		if( *p1 == '\t' ) {
			nxtpos = pos+8-pos%8;
			while(pos < trunc_len && pos <= nxtpos)
				newline[pos++] = ' ';
		} else if( *p1 != '\n' )
			newline[pos++]= *p1;
	}
	newline[pos] = '\0';
	return(newline);
}


/* ----------------------------------------------------------------------
 * dumplist:
 *	Outputs the current location counter, bytebuf[] array, and
 *	the string 'txt' to the listing file.
 *	This routine is called for every source line encountered in the
 *	source file. (Only in pass 2, and if listing is turned on).
 *
 */

void dumplist(char *txt, int show)
{
	int i,j;

	fprintf(listing,show?"%04X: ":"      ", lc);

	j = 0;
	for(i = 0; i < bytecount; i++ ) {
		fprintf(listing,"%02X ",bytebuf[i]);
		if( ++j >= 4 ) {
			j = 0;
			fprintf(listing,"\n      ");
		}
	}
	while(++j <= 4)
		fprintf(listing,"   ");

	fprintf(listing," %s\n",padline(txt));
}

/* ----------------------------------------------------------------------
 * gen* routines:
 *	Place information into the bytebuf[] array, and also
 *	call emitbyte with the byte.
 *
 */

void genbyte(int b)
{
  if (if_expr ^ in_else) {
    if( bytecount < sizeof(bytebuf) )
      bytebuf[bytecount++] = b;
    emitbyte(b);
  }
}

void genstr(const char *s)
{
  if (if_expr ^ in_else) {
	while( *s )
		genbyte(*s++);
  }
}

void genword(unsigned int w)
{
  if (if_expr ^ in_else) {
    /* "old" endian format
	genbyte( (w & 0xff00) >> 8 );
	genbyte( (w & 0x00ff) );
    */
    /* format of ASM504 */
	  genbyte( (w & 0x00ff) );
	  genbyte( (w & 0xff00) >> 8 );
  }
}
